chapter \<open>experiments on scanning and parsing\<close>

theory Test_Parsers
imports Main
begin

text \<open>
  experiments on scanning and parsing following Stefan Berghofers presentation
  at the Isabelle Developer Workshop 2010.
\<close>

text \<open>parse following Stefan Berghofer at Isabelle Developer Workshop 2010\<close>
ML \<open>
fun filtered_input str =
  filter Token.is_proper (Token.explode (Thy_Header.get_keywords' @{context}) Position.none str)
fun parse p str = Scan.finite Token.stopper p (filtered_input str);
\<close> ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close>

section \<open>scanner combinators from src/Pure/General/scan.ML\<close>
text \<open>ATTENTION, there are several similar definitions:
src> grep -r 'infix ' * | grep '\-\- ' | less
HOL/Import/proof_kernel.ML:    infix 7 |-- --|
HOL/Import/proof_kernel.ML:    infix 5 :-- -- ^^
HOL/Import/mono_scan.ML:infix 7 |-- --|
HOL/Import/mono_scan.ML:infix 5 :-- -- ^^
HOL/Import/scan.ML:infix 7 |-- --|
HOL/Import/scan.ML:infix 5 :-- -- ^^
Pure/General/scan.ML:infix 5 -- :-- :|-- |-- --| ^^;
\<close>

subsection \<open>one typical usage per combinator\<close>
text \<open>typical signatures\<close>
ML \<open>
parse;
(*fn : (Token.T list -> 'a * Token.T list) -> string -> 'a * Token.T list*)
      Parse.nat;
      (*fn : int Parse.parser*)
      (*fn : (Token.T list -> 'a * Token.T list)*)
parse Parse.nat;
(*fn : string -> int * Token.T list*)

parse Parse.nat "123";
(*(123, []) : int * Token.T list*)
\<close>

subsubsection \<open>op >>\<close>
ML \<open>
Parse.nat;
(fn i => i + 1);
op >>;
\<close> ML \<open>
(Parse.nat >> (fn i => i + 1));
parse;
\<close> ML \<open>
parse (Parse.nat >> (fn i => i + 1))
(*
parse (Parse.nat >> (fn i => i + 1));
       int Parse.parser
                     int -> int
                 ('a -> 'b * 'c) * ('b -> 'd) -> 'a -> 'd * 'c
      Token.T list -> int * Token.T list
(Token.T list -> 'a * Token.T list) -> string -> 'a * Token.T list
val it = fn : string -> int * Token.T list
*)
\<close>

subsubsection \<open>op ||\<close>
ML \<open>
parse (Args.$$$ "aaa" || Args.$$$ "bbb" || Args.$$$ "ccc") "aaa 111";
parse (Args.$$$ "aaa" || Args.$$$ "bbb" || Args.$$$ "ccc") "ccc 111";
(*val it = ("ccc", [Token (("111", ({}, {})), (Nat, "111"), Slot)])
   : string * Token.T list*)
parse ((Args.$$$ "aaa" -- Parse.nat) || (Args.$$$ "bbb" -- Parse.nat)) "aaa 111";
parse ((Args.$$$ "aaa" -- Parse.nat) || (Args.$$$ "bbb" -- Parse.nat)) "bbb 222";
\<close>

text \<open>|| requires arguments of equal type; 2 tricks that help:\<close>
ML \<open>
(* (1) works for prefixes complete enough to trigger a semantic action *)
fun exec1 str = "EXEC IMMEDIATELY: " ^ str;
fun exec2 (str, int) = "EXEC IMMEDIATELY: " ^ str ^ ", " ^ string_of_int int;
\<close> ML \<open>
(* (2) works for combinations of complete prefixes *)
fun drop_prefix _ = "PREFIX DROPPED";

Args.$$$ "aaa" >> exec1; 
Args.$$$ "bbb" -- Parse.nat >> exec2;
\<close> ML \<open>
(*+*)parse (Args.$$$ "aaa") "aaa";
(*+*)parse (Args.$$$ "aaa") "aaa xxx";
\<close> ML \<open>

Args.$$$ "aaa" >> exec1       || Args.$$$ "bbb" -- Parse.nat >> exec2;
Args.$$$ "aaa" >> drop_prefix || Args.$$$ "bbb" -- Parse.nat >> drop_prefix;
\<close> ML \<open>
(*+*)val p1 =
(*+*)  (Args.$$$ "aaa" >> exec1       || Args.$$$ "bbb" -- Parse.nat >> exec2): string parser;
(*+*)val p2 =
(*+*)  (Args.$$$ "aaa" >> drop_prefix || Args.$$$ "bbb" -- Parse.nat >> drop_prefix): string parser;
\<close> ML \<open>
parse p1 "aaa";     (* = ("EXEC IMMEDIATELY: aaa", [])*)
parse p1 "bbb 111"; (* = ("EXEC IMMEDIATELY: aaa", [])*)
\<close> ML \<open>
parse p2 "aaa";     (* = ("PREFIX DROPPED", [])*)
parse p2 "bbb 111"; (* = ("PREFIX DROPPED", [])*)
\<close> ML \<open>
\<close>


subsubsection \<open>op :--\<close>
ML \<open>(*from src/Pure/Isar/*)
            Parse.sym_ident : Token.T list -> string * Token.T list;
            (*fn : string Parse.parser*)
                                (fn "-" => Scan.succeed () | _ => Scan.fail);
                                (fn "-" => Scan.succeed ());
                                (*fn : string -> 'a -> unit * 'a*)
            Parse.sym_ident :-- (fn "-" => Scan.succeed () | _ => Scan.fail);
            (*fn : Token.T list -> (string * unit) * Token.T list*)
op >>;
(*fn : ('a -> 'b * 'c) * ('b -> 'd) -> 'a -> 'd * 'c*)

val minus = Parse.sym_ident :-- (fn "-" => Scan.succeed () | _ => Scan.fail) >> #1;
(*fn : Token.T list -> string * Token.T list*)
"------------------------------------------";
\<close> ML \<open>
parse minus "-aaa";
parse minus "- -";
(*WHY NOT SIMPLY Args.$$$ "-" ?*)
(*parse minus "111 -";*)
"------------------------------------------";
\<close> ML \<open>
infix 5 :--;
fun (scan1 -- scan2) = scan1 :-- (fn _ => scan2);
\<close>

subsubsection \<open>op --\<close>
ML \<open>
parse (minus -- Parse.nat)  "- 111";

Args.$$$ "theory";
Args.name;
(Args.$$$ "theory" -- Args.name);
parse (Args.$$$ "theory" -- Args.name);
(*
parse (Args.$$$ "theory")                GOON ?!?!?
  "theory";

parse (Args.$$$ "theory")
  "theory Sketch imports Main";
parse (Args.$$$ "theory" -- Args.name)
  "theory Sketch imports Main";
parse (Args.$$$ "theory" -- Args.name -- Args.$$$ "imports" -- Args.name)
  "theory Sketch imports Main";
*)
\<close>

subsubsection \<open>op :|-- TODO\<close>
ML \<open>
fun token atom = Scan.ahead Parse.not_eof --| atom;
(*fn : (Token.T list -> 'a * 'b) -> Token.T list -> Token.T * 'b*)

val ident = token
  (Parse.short_ident || Parse.long_ident || Parse.sym_ident || Parse.term_var ||
    Parse.type_ident || Parse.type_var || Parse.number);
(*fn : Token.T list -> Token.T * Token.T list*)

(ident >> Token.content_of) :|-- (fn x =>
  (case Lexicon.read_variable x of SOME v => Scan.succeed v | NONE => Scan.fail));
(*fn : Token.T list -> indexname * Token.T list*)
\<close>

subsubsection \<open>op |--\<close>
ML \<open>(*from Pure/Isar/args.ML, fun attribs*)
parse (Parse.$$$ "[" |-- Parse.list1 Parse.name) "[111, bbb, ccc]";
\<close>

subsubsection \<open>op --|\<close>
ML \<open>(*from Pure/Isar/args.ML, fun attribs*)
parse (Parse.list1 Parse.name --| Parse.$$$ "]") " 111, bbb, ccc]";
\<close>

section \<open>parse compound suffixes\<close>
subsection \<open>Lists\<close>
ML \<open>(*from Pure/Isar/args.ML, fun attribs*)
Parse.$$$ "[" |-- Parse.list1 Parse.name --| Parse.$$$ "]";
(*fn : Token.T list -> bstring list * Token.T list*)
\<close> ML \<open>
\<close> ML \<open> (* example close to lists see Cookbook p.144 \<section>6.2 *)
let
  val input = filtered_input "in , in , in"
  val p = Parse.enum "," (Parse.$$$ "in")
in
  Scan.finite Token.stopper p input
end
(* = (["in", "in", "in"], []) *)
\<close> ML \<open>
let
  val input = filtered_input "aaa , bbb , ccc"
  val p = Parse.enum "," Parse.name
in
  Scan.finite Token.stopper p input
end
(* = (["aaa", "bbb", "ccc"], []) *)
\<close> ML \<open>
let
  val input = filtered_input "\"aaa\" , \"bbb\" , \"ccc\""
  val p = Parse.enum "," Parse.string
in
  Scan.finite Token.stopper p input
end
(* = (["aaa", "bbb", "ccc"], []) *)
\<close> ML \<open>
let
  val input = filtered_input "\"aaa\" , \"bbb\" , \"ccc\""
  val p = Parse.list1 Parse.string
in
  Scan.finite Token.stopper p input
end
(* = (["aaa", "bbb", "ccc"], []) *)
\<close> ML \<open>
let
  val input = filtered_input "[\"aaa\" , \"bbb\" , \"ccc\"]"
  val p = Parse.$$$ "[" |-- Parse.list1 Parse.string --| Parse.$$$ "]"
in
  Scan.finite Token.stopper p input
end
(* = (["aaa", "bbb", "ccc"], []) *)



\<close> ML \<open>
parse (Parse.$$$ "[" |-- Parse.list Parse.name --| Parse.$$$ "]")
  "[aaa, bbb, ccc]"; 
\<close> ML \<open>
parse ((Parse.$$$ "[" |-- Parse.list Parse.int --| Parse.$$$ "]") --
       (Parse.$$$ "[" |-- Parse.list Parse.name --| Parse.$$$ "]"))
  "[123, 456, 789] [aaa, bbb, ccc]";
(* = (([123, 456, 789], ["aaa", "bbb", "ccc"]), [])*)
\<close> ML \<open>
\<close> ML \<open>
\<close>

subsection \<open>terms\<close>
text \<open>see src/Pure/isac/smltest/Pure/Isar/Test_Parse_Term.thy\<close>

subsection \<open> -- Scan.optional\<close>
subsubsection \<open>src/HOL/Tools/value_command.ML ..Scan.optional\<close>
text \<open>
  this works nicely ...
  but for ISAC we need: problem_headline -- Scan.optional (...)
\<close>
ML \<open>
\<close> ML \<open>
val opt_modes =
  Scan.optional (\<^keyword>\<open>(\<close> |-- Parse.!!! (Scan.repeat1 Parse.name --| \<^keyword>\<open>)\<close>)) ([]: string list)
:                                                                                  (*^^^^^^^1^^^^^^^^*)
  Token.T list -> string list * Token.T list
\<close> ML \<open>
(\<^keyword>\<open>(\<close> |-- Parse.!!! (Scan.repeat1 Parse.name --| \<^keyword>\<open>)\<close>)): string list parser
                                                                    (*^^^^^1^^^^^*)
\<close> ML \<open>
val opt_evaluator =
  Scan.optional (\<^keyword>\<open>[\<close> |-- Parse.name --| \<^keyword>\<open>]\<close>) ("": string)
:                                                         (*^^^^2^^^^^^*)
  Token.T list -> string * Token.T list
\<close> ML \<open>
(\<^keyword>\<open>[\<close> |-- Parse.name --| \<^keyword>\<open>]\<close>): string parser
                                           (*^^2^^*)
\<close> ML \<open>
val parse_value = opt_evaluator -- opt_modes -- Parse.term
:
Token.T list -> ((string * string list) * string) * Token.T list
(*                eval     modes          term*)
\<close> ML \<open>
\<close>
(*/----- requires HOL.Complex --------------\
value "1 + 2 = (3::int)" (* <---------------- *)
  \----- requires HOL.Complex --------------/*)
ML \<open>
val toks = filtered_input "\"1 + 2 = (3::int)\""
\<close> ML \<open>
parse_value toks (* = ((("", []), "<markup>"), [])*)
\<close>
(*/----- requires HOL.Complex --------------\
value [simp] "1 + 2 = (3::int)" (* <---------------- *)
  \----- requires HOL.Complex --------------/*)
ML \<open>
val toks = filtered_input "[simp] \"1 + 2 = (3::int)\""
\<close> ML \<open>
parse_value toks (* = ((("simp", []), "<markup>"), [])*)
\<close>
(*/----- requires HOL.Complex --------------\
value [simp] (asci latex) "1 + 2 = (3::int)" (* <---------------- *)
  \----- requires HOL.Complex --------------/*)
ML \<open>
val toks = filtered_input "[simp] (asci latex) \"1 + 2 = (3::int)\""
\<close> ML \<open>
parse_value toks (* = ((("simp", ["asci", "latex"]), "<markup>"), [])*)
\<close>

subsubsection \<open>cp src/../parse.ML\<close>
text \<open>
  this is more complicated than Cookbook Scan.optional p.146 \<section>6.7
  thus stopped.
\<close>
ML \<open>
\<close> ML \<open> (*cp src/../parse.ML: val is_terms = Scan.repeat1 ($$$ "is" |-- term); etc*)
val is_terms = Scan.repeat1 (Parse.$$$ "is" |-- Parse.term);
val termp = Parse.term --
  Scan.optional (Parse.$$$ "(" |-- Parse.!!! (is_terms --| Parse.$$$ ")")) ([]: string list);
                                                                              (*^^^^^^^^^^^*)
\<close> ML \<open>
termp: (string * string list) parser;
\<close> ML \<open>
(Parse.$$$ "(" |-- Parse.!!! (is_terms --| Parse.$$$ ")")) : string list parser
                                                           (*^^^^^^^^^^^*)
\<close> ML \<open>
parse termp "\"aaa + bbb\" (is \"aaa + bbb\") ";
(* = (("<markup>", ["<markup>"]), []) *)
\<close> ML \<open>
parse termp "\"aaa + bbb\" (is \"aaa + bbb\" is \"123 + 456\") ";
(* = (("<markup>", ["<markup>", "<markup>"]), []) *)
\<close> ML \<open>
\<close> ML \<open>
\<close>

subsubsection \<open>..simplify towards ISAC 1\<close>
ML \<open>
val is_terms = Scan.repeat1 ((*Parse.$$$ "is" |--*) Parse.name) : string list parser
\<close> ML \<open>
val termp =
  Parse.reserved "xxx" --
    Scan.optional
      (Parse.$$$ "(" |-- Parse.!!! (is_terms --| Parse.$$$ ")"))
      ([](*!..*): string list)(*..!*);
\<close> ML \<open>
  Parse.term : string parser
\<close> ML \<open>
    Scan.optional: ('a -> 'b * 'a) -> 'b -> 'a -> 'b * 'a
    (*these MUST be same: ^^          []          ^^               string list *)
\<close> ML \<open>
    Scan.optional Parse.int: int -> Token.T list -> int * Token.T list;
    Scan.optional Parse.int: int -> int parser;
\<close> ML \<open>
      (Parse.$$$ "(" |-- Parse.!!! (is_terms --| Parse.$$$ ")")) : string list parser
\<close> ML \<open>
parse termp "xxx";          (* = (("xxx", []), [])*)
parse termp "xxx ( name )"; (* = (("xxx", ["name"]), [])*)
\<close> ML \<open>
\<close>

subsubsection \<open>..simplify towards ISAC 2\<close>
ML \<open>
     Parse.reserved "Specification" -- Parse.$$$ ":" --
(**)   Scan.optional (**)
         (Parse.list Parse.name)
(**)     ([] : string list) (**)
\<close> ML \<open>
       Scan.optional: ('a -> 'b * 'a) -> 'b -> 'a -> 'b * 'a
       (*these MUST be same: ^^          []          ^^       string list *)
\<close> ML \<open>
\<close>

subsubsection \<open>cp src/../parse_spec.ML\<close>
ML \<open>
\<close> ML \<open> (*cp val locale_expression = *)
val instance = Scan.optional (Parse.where_ |--
  Parse.and_list1 (Parse.name -- (Parse.$$$ "=" |-- Parse.term)) >> Expression.Named ||
  Scan.repeat1 (Parse.maybe Parse.term) >> Expression.Positional) (Expression.Named []) --
  (Scan.optional (Parse.$$$ "rewrites" |-- Parse.and_list1 (Parse_Spec.opt_thm_name ":" -- Parse.prop)) []);
fun plus1_unless test scan =
  scan ::: Scan.repeat (Parse.$$$ "+" |-- Scan.unless test (Parse.!!! scan));
(*val locale_expression =
  let..*)
    val expr2 = Parse.name_position;
    val expr1 =
      Parse_Spec.locale_prefix -- expr2 --
      instance >> (fn ((p, l), i) => (l, (p, i)));
    val expr0 = plus1_unless Parse_Spec.locale_keyword expr1;
\<close> ML \<open>
(*in..*)
expr0 -- Scan.optional
  (Parse.$$$ "for" |-- Parse.!!! Parse_Spec.locale_fixes)
  ([] : (binding * string option * mixfix) list)
       (*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*)
:
(((string * Position.T) * ((string * bool) * (string Expression.map * (Attrib.binding * string) list))) list *
    (binding * string option * mixfix) list) 
  parser
\<close> ML \<open>
expr0 :
 ((string * Position.T) * ((string * bool) * (string Expression.map * (Attrib.binding * string) list))) list 
  parser
\<close> ML \<open>
  (Parse.$$$ "for" |-- Parse.!!! Parse_Spec.locale_fixes)
:
    (binding * string option * mixfix) list parser
  (*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*)
\<close> ML \<open>
\<close>

subsection \<open>option and repetition\<close>
ML \<open>
val xxx = (Args.$$$ "top" -- (Scan.option (Args.$$$ "equiv"))
       -- Args.$$$ "dots");
\<close> ML \<open>
parse xxx "top dots";
parse xxx "top equiv dots";

val yyy = Scan.repeat1 (Args.$$$ "equiv");
parse yyy "equiv equiv equiv";
parse yyy "equiv";
(*
parse yyy "";
*** exception FAIL (SOME "keyword expected,\nbut end-of-file was found") raised (line 135 of "General/scan.ML")*)
\<close>

subsection \<open>recursive parsing\<close>
ML \<open>
fun body s xs = (Args.$$$ "top" >> drop_prefix 
             || (Args.$$$ "bullet" -- (body s)) >> drop_prefix) xs;
parse (body "top") "top";
parse (body "top") "bullet top";
parse (body "top") "bullet bullet bullet top";
"........ parsed all, however: ..........";

parse (body "top") "top bullet top";
(*val it =
   ("PREFIX DROPPED",
    [Token (("bullet", ({}, {})), (Ident, "bullet"), Slot),
     Token (("top", ({}, {})), (..., "top"), Slot)])
   : string * Token.T list*)
"-------- fst argument seems irrelavant ? ...";
parse (body "") "top";
parse (body "") "bullet top";
parse (body "") "bullet bullet bullet top";
parse (body "") "top bullet top";
\<close>

section \<open>GOON\<close>
subsection \<open>GOON\<close>


end