(* Title: tests for the interface
   Author: Mathias Lehnfeld 2013
   Use is subject to license terms.
*)

"-----------------------------------------------------------------------------------------------";
"table of contents -----------------------------------------------------------------------------";
"-----------------------------------------------------------------------------------------------";
"-----------------------------------------------------------------------------------------------";
"-------- test_applyFormula --------------------------------------------------------------------";
"-------- ME_Misc.get_interval from ctree with move_dn:modspec.sml -------------------";
"-----------------------------------------------------------------------------";
"-----------------------------------------------------------------------------";


"-------- test_applyFormula --------------------------------------------------------------------";
"-------- test_applyFormula --------------------------------------------------------------------";
"-------- test_applyFormula --------------------------------------------------------------------";
"----- this shows the principle how Isac's math-engine is made parallel ------";
(* a store of simplified states *)
val test_states = Unsynchronized.ref ([]: (int * string list) list)

(* a simplified getter *)
fun test_get_calc cI =
  case LibraryC.assoc (!test_states, cI) of
    NONE => error ("get_calc " ^ string_of_int cI ^ " not existent")
  | SOME c => c;

(* a simplified updater *)
fun test_upd_calc cI cs =
  (if is_none (LibraryC.assoc (!test_states, cI))
  then writeln ("upd_calc created new calculation " ^ string_of_int cI)
  else ();
  test_states := overwrite (!test_states, (cI, cs)))

(* a dummy call of the math-engine *)
fun test_math_engine cs = cs |> split_last ||> (fn x => x ^ "-CHECKED") ||> single |-> (curry op @);

test_states :=
  [(1, ["calc1-line1", "calc1-line2", "calc1-line3"]),
  (2, ["calc2-line1"]),
  (3, ["calc3-line1", "calc3-line2"])];
test_states := [];

val cI = 123;
test_upd_calc cI ["initialize 123"];
appendFormula cI;

"-------- ME_Misc.get_interval from ctree with move_dn:modspec.sml -------------------";
"-------- ME_Misc.get_interval from ctree with move_dn:modspec.sml -------------------";
"-------- ME_Misc.get_interval from ctree with move_dn:modspec.sml -------------------";
"=====new ctree 1: crippled by cut_level_'_ ======================";
reset_states ();
CalcTree
[(["equality (x/(x^2 - 6*x+9) - 1/(x^2 - 3*x) =1/x)",
	   "solveFor x", "solutions L"], 
  ("RatEq",["univariate", "equation"],["no_met"]))];
Iterator 1; moveActiveRoot 1;
autoCalculate 1 CompleteCalc; 

getTactic 1 ([1],Res);(*Rewrite_Set RatEq_simplify*)
getTactic 1 ([2],Res);(*Rewrite_Set norm_Rational*)
getTactic 1 ([3],Res);(*Rewrite_Set RatEq_eliminate*)
getTactic 1 ([4,1],Res);(*Rewrite all_left*)
getTactic 1 ([4,2],Res);(*Rewrite_Set expand_binoms*)
getTactic 1 ([4,3],Res);(*Rewrite_Set_Inst make_ratpoly_in*)

moveActiveFormula 1 ([1],Res)(*1.1...1.4*);
moveActiveFormula 1 ([2],Res)(**ME_Isa: 'expand' not known*);
moveActiveFormula 1 ([3],Res)(*3.1.*);
moveActiveFormula 1 ([4,2],Res)(*4.2.1.*);
moveActiveFormula 1 ([4,3],Res)(**one_scr_arg: called by Program Stepwise t_=*);

moveActiveFormula 1 ([1],Res)(*1.1...1.4*);
interSteps 1 ([1],Res)(*..is activeFormula !?!*);

getTactic 1 ([1,1],Res);(*Rewrite real_diff_minus*)
getTactic 1 ([1,2],Res);(*Rewrite real_diff_minus*)
getTactic 1 ([1,3],Res);(*Rewrite real_diff_minus*)
getTactic 1 ([1,4],Res);(*Rewrite real_rat_mult_1*)

moveActiveFormula 1 ([4,2],Res)(*4.2.1.*);
interSteps 1 ([4,2],Res)(*..is activeFormula !?!*);
val ((pt,_),_) = get_calc 1;
writeln(pr_ctree pr_short pt);
(*delete [4,1] in order to make pos [4],[4,4] for pblobjs differen [4],[4,3]:
 ###########################################################################*)
val (pt, _) = cut_level_'_ [] [] pt ([4,1],Frm);                         (*#*)
(*##########################################################################*)
writeln(pr_ctree pr_short pt);
(*##########################################################################
  attention: the ctree in states is still the old (perfect) !!!
############################################################################*)

writeln(pr_ctree pr_short pt);
writeln(Test_Tool.posterms2str (ME_Misc.get_interval ([],Frm) ([],Res) 99999 pt));

case map fst3 (ME_Misc.get_interval ([],Frm) ([],Res) 99999 pt) of
    [([], Frm), 
     ([1], Frm), 
     ([1, 1], Frm), 
     ([1, 1], Res), 
     ([1, 2], Res),
     ([1, 3], Res), 
     ([1, 4], Res), 
     ([1], Res), 
     ([2], Res), 
     ([3], Res),
     ([4], Pbl), 
     ([4, 1], Frm), 
     ([4, 1, 1], Frm), 
     ([4, 1, 1], Res),
     ([4, 1], Res), 
     ([4, 2], Res), 
     ([4, 3], Pbl), 
     ([4, 3, 1], Frm),
     ([4, 3, 1], Res), 
     ([4, 3, 2], Res), 
     ([4, 3, 3], Res), 
     ([4, 3, 4], Res),
     ([4, 3, 5], Res), 
     ([4, 3], Res), 
     ([4], Res), 
     ([5], Res), 
     ([], Res)] => () 
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1f";
case map fst3 (ME_Misc.get_interval ([],Frm) ([],Res) 1 pt) of
    [([], Frm), 
     ([1], Frm), 
     ([1], Res), 
     ([2], Res), 
     ([3], Res),
     ([4], Pbl), 
     ([4], Res), 
     ([5], Res), 
     ([], Res)] => () 
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1f";
case map fst3 (ME_Misc.get_interval ([],Frm) ([],Res) 0 pt) of
    [([], Frm), 
     ([], Res)] => () 
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1f";

case map fst3 (ME_Misc.get_interval ([1,3],Res) ([4,1,1],Frm) 99999 pt) of
    [([1, 3], Res), 
     ([1, 4], Res), 
     ([1], Res), 
     ([2], Res), 
     ([3], Res),
     ([4], Pbl), 
     ([4, 1], Frm), 
     ([4, 1, 1], Frm)] => () 
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1a";

(*this pos' is not generated bu move_dn:......vvv: goes to end of calc*)
case map fst3 (ME_Misc.get_interval ([2],Res) ([4,3,2],Frm) 99999 pt) of
    [([2], Res), 
     ([3], Res), 
     ([4], Pbl), 
     ([4, 1], Frm), 
     ([4, 1, 1], Frm),
     ([4, 1, 1], Res), 
     ([4, 1], Res), 
     ([4, 2], Res), 
     ([4, 3], Pbl),
     ([4, 3, 1], Frm), 
     ([4, 3, 1], Res), 
     ([4, 3, 2], Res), 
     ([4, 3, 3], Res),(*this is beyond 'to'*)
     ([4, 3, 4], Res),(*this is beyond 'to'*) 
     ([4, 3, 5], Res),(*this is beyond 'to'*) 
     ([4, 3], Res),   (*this is beyond 'to'*)
     ([4], Res),      (*this is beyond 'to'*)
     ([5], Res),      (*this is beyond 'to'*)
     ([], Res)] => () (*this is beyond 'to'*)
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1b";
case map fst3 (ME_Misc.get_interval ([1,4],Res) ([4,3,1],Frm) 99999 pt) of
    [([1, 4], Res), 
     ([1], Res), 
     ([2], Res), 
     ([3], Res), 
     ([4], Pbl),
     ([4, 1], Frm), 
     ([4, 1, 1], Frm), 
     ([4, 1, 1], Res), 
     ([4, 1], Res),
     ([4, 2], Res), 
     ([4, 3], Pbl), 
     ([4, 3, 1], Frm)] => () 
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1c";
case map fst3 (ME_Misc.get_interval ([4,2],Res) ([5],Res) 99999 pt) of
    [([4, 2], Res), 
     ([4, 3], Pbl), 
     ([4, 3, 1], Frm), 
     ([4, 3, 1], Res),
     ([4, 3, 2], Res), 
     ([4, 3, 3], Res), 
     ([4, 3, 4], Res), 
     ([4, 3, 5], Res),
     ([4, 3], Res), 
     ([4], Res), 
     ([5], Res)]=>()
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1d";
case map fst3 (ME_Misc.get_interval ([],Frm) ([4,3,2],Res) 99999 pt) of
    [([], Frm), 
     ([1], Frm), 
     ([1, 1], Frm), 
     ([1, 1], Res), 
     ([1, 2], Res),
     ([1, 3], Res), 
     ([1, 4], Res), 
     ([1], Res), 
     ([2], Res), 
     ([3], Res),
     ([4], Pbl), 
     ([4, 1], Frm), 
     ([4, 1, 1], Frm), 
     ([4, 1, 1], Res),
     ([4, 1], Res), 
     ([4, 2], Res), 
     ([4, 3], Pbl), 
     ([4, 3, 1], Frm),
     ([4, 3, 1], Res), 
     ([4, 3, 2], Res)] => () 
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1e";
case map fst3 (ME_Misc.get_interval ([4,3],Frm) ([4,3],Res) 99999 pt) of
    [([4, 3], Frm), 
     ([4, 3, 1], Frm), 
     ([4, 3, 1], Res), 
     ([4, 3, 2], Res),
     ([4, 3, 3], Res), 
     ([4, 3, 4], Res), 
     ([4, 3, 5], Res), 
     ([4, 3], Res)] => () 
  | _ => error "diff.behav.in ctree.sml: ME_Misc.get_interval lev 1g";
